"""Decorators and helpers to check access to communities."""

from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from typing import Any

from flask import g
from flask_login import current_user
from werkzeug.exceptions import Forbidden
from werkzeug.wrappers import Response

from abilian.core.models.subjects import User
from abilian.sbe.apps.communities.presenters import CommunityPresenter
from abilian.services import security_service
from abilian.services.security import MANAGE


def require_admin(func: Callable) -> Callable:
    @wraps(func)
    def decorated_view(*args: Any, **kwargs: Any) -> str | Response:
        is_admin = security_service.has_role(current_user, "admin")
        if not is_admin:
            raise Forbidden
        return func(*args, **kwargs)

    return decorated_view


def require_manage(func: Callable) -> Callable:
    @wraps(func)
    def decorated_view(*args: Any, **kwargs: Any) -> Response:
        community = g.community
        if community and community.has_permission(current_user, MANAGE):
            return func(*args, **kwargs)
        is_admin = security_service.has_role(current_user, "admin")
        if not is_admin:
            raise Forbidden
        return func(*args, **kwargs)

    return decorated_view


def require_access(func):
    @wraps(func)
    def decorated_view(*args, **kwargs):
        check_access()
        return func(*args, **kwargs)

    return decorated_view


def check_access(
    community: CommunityPresenter | None = None, user: User | None = None
) -> None:
    if not has_access(community, user):
        raise Forbidden


def has_access(
    community: CommunityPresenter | None = None, user: User | None = None
) -> bool:
    if not user:
        user = current_user
    if user.is_anonymous:
        return False

    is_admin = security_service.has_role(user, "admin")
    if is_admin:
        return True

    if not community:
        community = getattr(g, "community", None)

    if community is not None:
        return community.get_role(user) is not None

    return False


def is_manager(context: dict[str, Any] | None = None, user: User | None = None) -> bool:
    if not user:
        user = current_user
    if user.is_anonymous:
        return False

    if context:
        community = context.get("object").community  # type: ignore
    else:
        community = g.community

    if community.has_permission(user, MANAGE) or user == community.creator:
        return True

    return security_service.has_role(user, "admin")



#!/usr/bin/env python
"""Read public domain (CC-0) database of IP/country and export to MMDB format.

Resulting file file is written in curent working directory.
"""
import csv
import ssl
import urllib.request
from collections import defaultdict
from pathlib import Path
from tempfile import TemporaryDirectory

import certifi
import maxminddb
from mmdb_writer import MMDBWriter
from netaddr import IPRange, IPSet

CSV_DB = (
    "https://cdn.jsdelivr.net/npm/@ip-location-db/geo-whois-asn-country/"
    "geo-whois-asn-country-ipv4.csv"
)
MMDB_FILENAME = "ip_country.mmdb"


def fetch_csv_db(csv_db_path: Path) -> None:
    with urllib.request.urlopen(  # noqa: S310
        CSV_DB, context=ssl.create_default_context(cafile=certifi.where())
    ) as response:
        csv_db_path.write_bytes(response.read())
    size = csv_db_path.stat().st_size / 2**20
    print(f"CSV DB: {csv_db_path.name} {size:.1f} MB")


def write_db(db_path: Path, csv_db_path: Path):
    data = defaultdict(list)
    with open(csv_db_path) as file:
        reader = csv.reader(file)
        for line in reader:
            data[line[2]].append(IPRange(line[0], line[1]))

    writer = MMDBWriter()
    for country, cidrs in data.items():
        writer.insert_network(IPSet(cidrs), {"country": country})
    writer.to_db_file(str(db_path))


def check_db_done(db_path: Path) -> None:
    free_fr_ip = "212.27.48.10"
    ibm_com_ip = "104.85.45.20"
    db = maxminddb.open_database(db_path)
    result1 = db.get(free_fr_ip)
    assert result1["country"] == "FR"
    result2 = db.get(ibm_com_ip)
    assert result2["country"] == "US"


def main() -> None:
    local_path = Path.cwd()
    db_path = local_path / MMDB_FILENAME
    with TemporaryDirectory(prefix="ipcountry-", dir=local_path) as temp_dir:
        print(f"temp dir: {temp_dir}")
        csv_db_path = Path(temp_dir) / "ip_country.csv"
        fetch_csv_db(csv_db_path)
        write_db(db_path, csv_db_path)
    check_db_done(db_path)
    size = db_path.stat().st_size // 2**20
    print(f"MMDB: {db_path.name} {size:.1f} MB")


if __name__ == "__main__":
    main()



"""Communities module."""

from __future__ import annotations

from flask import Flask


def register_plugin(app: Flask) -> None:
    # Used for side effects
    from . import events, search
    from .views import communities

    app.register_blueprint(communities)

    search.init_app(app)



from __future__ import annotations

import os

import click
import toml
from dotenv import load_dotenv
from flask import Blueprint, Response, abort, current_app, g, redirect, request
from flask.cli import AppGroup, FlaskGroup
from flask_login import current_user
from loguru import logger

import abilian.cli
from abilian.app import Application as BaseApplication
from abilian.core.dramatiq.setup import init_dramatiq_engine
from abilian.core.logging import init_logging
from abilian.i18n import _l
from abilian.sbe.apps.social.views.social import social
from abilian.sbe.extension import sbe
from abilian.web.action import actions
from abilian.web.nav import NavItem
from abilian.web.util import url_for

from .config import BaseConfig

__all__ = ["create_app"]

HOME_ACTION = NavItem("section", "home", title=_l("Home"), endpoint="social.home")


class Application(BaseApplication):
    APP_PLUGINS = BaseApplication.APP_PLUGINS + [
        "abilian.sbe.apps.notifications",
        "abilian.sbe.apps.wiki",
        "abilian.sbe.apps.wall",
        "abilian.sbe.apps.documents",
        "abilian.sbe.apps.forum",
        "abilian.sbe.apps.communities",
        "abilian.sbe.apps.social",
        "abilian.sbe.apps.preferences",
    ]

    def init_extensions(self):
        super().init_extensions()

        sbe.init_app(self)


def create_app(config=None, **kw):
    app = Application(__name__, **kw)

    app.config.from_object(BaseConfig)

    if config:
        app.config.from_object(config)

    if not config:
        read_config(app)
        app.config.from_prefixed_env()

    # Setup stuff.
    init_logging(app)

    if "DATABASE_URL" in os.environ:
        app.config["SQLALCHEMY_DATABASE_URI"] = os.environ["DATABASE_URL"]
    # We must register this before blueprint is registered
    if not any(
        url
        for url in social.url_value_preprocessors.get(None, [])
        if url.__name__ == "on_home_blueprint"
    ):
        if app.config["TESTING"]:
            social._got_registered_once = False
        social.url_value_preprocessor(on_home_blueprint)

    with app.app_context():
        app.setup(config)

    with app.app_context():
        actions.register(HOME_ACTION)

    app.register_blueprint(MAIN)
    app.before_request(login_required)

    app.register_error_handler(400, csrf_error_response)

    register_cli(app)

    init_dramatiq_engine(app)

    # Done
    return app


def read_config(app):
    load_dotenv()

    env = os.environ.get("FLASK_CONFIG", "development")

    read_toml_config(app, "base.toml")
    if env == "development":
        read_toml_config(app, "development.toml")
    else:
        read_toml_config(app, "production.toml")

    read_toml_config(app, "secrets.toml")


def read_toml_config(app, filename):
    absolute_filename = os.path.join(os.getcwd(), "config", filename)
    if os.path.exists(absolute_filename):
        app.config.from_file(absolute_filename, load=toml.load)


@click.group(cls=FlaskGroup, create_app=create_app)
def cli():
    """Management script for the labster application."""


def register_cli(app):
    for obj in vars(abilian.cli).values():
        if isinstance(obj, (click.core.Command, AppGroup)):
            app.cli.add_command(obj)


def csrf_error_response(reason):
    # let sentry be aware of csrf failures. They might show app is broken
    # somewhere
    logger.error(
        "Csrf error report, reason: {reason}",
        reason=reason,
        extra={"stack": True},
    )
    return abort(400, reason)


MAIN = Blueprint("main", __name__, url_prefix="")


@MAIN.route("/")
def home():
    """
    Home page. Actually there is no home page, so we redirect to the most
    appropriate place.
    """
    return redirect(url_for("social.home"))


def login_required() -> Response | None:
    """
    Before request handler to ensure login is required on any view
    """
    if current_app.config.get("NO_LOGIN"):
        return None

    if request.path.startswith(current_app.static_url_path):
        return None

    if request.url_rule and request.url_rule.endpoint == "geodata.static":
        # geodata blueprint assets
        return None

    if request.blueprint in ("login", "first_login", "projects", "notifications"):
        return None

    if request.blueprint == "_debug_toolbar" and current_app.debug:
        return None

    if not current_user.is_authenticated:
        return redirect(url_for("login.login_form", next=request.url))

    return None


def on_home_blueprint(endpoint, values):
    """
    url_value_preprocessor attached to blueprint used for 'home'
    """
    if endpoint == HOME_ACTION.endpoint.name:
        g.nav["active"] = HOME_ACTION.path



""""""

from __future__ import annotations

import os

import click
from flask import current_app
from flask.cli import with_appcontext


@click.command(short_help="Show config")
@with_appcontext
def config() -> None:
    config_ = dict(sorted(current_app.config.items()))
    print("CONFIG:\n")
    for k, v in config_.items():
        if not k[0].isupper():
            continue
        try:
            print(f"{k}: {v}")
        except Exception as e:
            print(f"{k}: {e}")

    print("\nENV:\n")

    env_ = dict(sorted(os.environ.items()))
    for k, v in env_.items():
        print(f"{k}: {v}")



"""Calendar module."""

from __future__ import annotations


def register_plugin(app):
    from .actions import register_actions
    from .views import blueprint

    blueprint.record_once(register_actions)
    app.register_blueprint(blueprint)



""""""

from __future__ import annotations

from typing import Any

from flask import g
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey
from whoosh.fields import STORED

from .base import AUDITABLE, EDITABLE, SEARCHABLE, SYSTEM
from .subjects import User


class OwnedMixin:
    __index_to__ = (
        ("creator", ("creator",)),
        ("creator_name", (("creator_name", STORED),)),
        ("owner", ("owner",)),
        ("owner_name", (("owner_name", STORED),)),
    )

    def __init__(self, *args: list, **kwargs: dict[str, Any]):
        try:
            user = g.user
            if not self.creator and not g.user.is_anonymous:
                self.creator = user
            if not self.owner and not g.user.is_anonymous:
                self.owner = user
        except (RuntimeError, AttributeError):
            pass

    @declared_attr
    def creator_id(cls):
        return Column(ForeignKey(User.id), info=SYSTEM)

    @declared_attr
    def creator(cls):
        primary_join = f"User.id == {cls.__name__}.creator_id"
        return relationship(
            User,
            primaryjoin=primary_join,
            lazy="joined",
            uselist=False,
            info=SYSTEM | SEARCHABLE,
        )

    @property
    def creator_name(self) -> str:
        return str(self.creator) if self.creator else ""

    @declared_attr
    def owner_id(cls):
        return Column(ForeignKey(User.id), info=EDITABLE | AUDITABLE)

    @declared_attr
    def owner(cls):
        primary_join = f"User.id == {cls.__name__}.owner_id"
        return relationship(
            User,
            primaryjoin=primary_join,
            lazy="joined",
            uselist=False,
            info=EDITABLE | AUDITABLE | SEARCHABLE,
        )

    @property
    def owner_name(self) -> str:
        return str(self.owner) if self.owner else ""



"""Security service, manages roles and permissions."""

from __future__ import annotations

from collections.abc import Callable, Collection
from functools import wraps
from itertools import chain
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
from flask import g
from flask_login import current_user
from sqlalchemy import sql
from sqlalchemy.orm import Session, object_session, subqueryload
from sqlalchemy.sql.selectable import Exists

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.core.extensions.login import AnonymousUser
from abilian.core.models import Model
from abilian.core.models.subjects import Group, Principal, User
from abilian.core.util import unwrap
from abilian.services import Service, ServiceState
from abilian.services.security.models import (
    CREATE,
    DELETE,
    MANAGE,
    PERMISSIONS_ATTR,
    READ,
    WRITE,
    Admin,
)
from abilian.services.security.models import Anonymous as AnonymousRole
from abilian.services.security.models import (
    Authenticated,
    Creator,
    FolderishModel,
    InheritSecurity,
    Manager,
    Owner,
    Permission,
    PermissionAssignment,
    Reader,
    Role,
    RoleAssignment,
    SecurityAudit,
    Writer,
)

if TYPE_CHECKING:
    from abilian.app import Application


#: list of legacy supported permissions when not using :class:`Permission`
#: instance
PERMISSIONS = frozenset(["read", "write", "manage"])

__all__ = [
    "InheritSecurity",
    "SecurityAudit",
    "SecurityError",
    "SecurityService",
    "security",
]

#: default security matrix
DEFAULT_PERMISSION_ROLE: dict[Permission, frozenset[Role]] = {}
prm = DEFAULT_PERMISSION_ROLE
prm[MANAGE] = frozenset({Admin, Manager})
prm[WRITE] = frozenset({Admin, Manager, Writer})
prm[CREATE] = frozenset({Admin, Manager, Writer})
prm[DELETE] = frozenset({Admin, Manager, Writer})
prm[READ] = frozenset({Admin, Manager, Writer, Reader})
del prm


class SecurityError(Exception):
    pass


class SecurityServiceState(ServiceState):
    use_cache = True
    #: True if security has changed
    needs_db_flush = False


def require_flush(fun: Callable) -> Callable:
    """Decorator for methods that need to query security.

    It ensures all security related operations are flushed to DB, but
    avoids unneeded flushes.
    """

    @wraps(fun)
    def ensure_flushed(service: SecurityService, *args: Any, **kwargs: Any) -> Any:
        if service.app_state.needs_db_flush:
            session = db.session()
            if not session._flushing and any(
                isinstance(m, (RoleAssignment, SecurityAudit))
                for models in (session.new, session.dirty, session.deleted)
                for m in models
            ):
                session.flush()
            service.app_state.needs_db_flush = False

        return fun(service, *args, **kwargs)

    return ensure_flushed


def query_pa_no_flush(
    session: Session, permission: Permission, role: Role, obj: Model | None
):
    """Query for a :class:`PermissionAssignment` using `session` without any
    `flush()`.

    It works by looking in session `new`, `dirty` and `deleted`, and issuing a
    query with no autoflush.

    .. note::

        This function is used by `add_permission` and `delete_permission` to allow
        to add/remove the same assignment twice without issuing any flush. Since
        :class:`Entity` creates its initial permissions in during
        :sqlalchemy:`sqlalchemy.orm.events.SessionEvents.after_attach`, it might be
        problematic to issue a flush when entity is not yet ready to be flushed
        (missing required attributes for example).
    """
    to_visit = [session.deleted, session.dirty, session.new]
    with session.no_autoflush:
        # no_autoflush is required to visit PERMISSIONS_ATTR without emitting a
        # flush()
        if obj:
            to_visit.append(getattr(obj, PERMISSIONS_ATTR))

        permissions = (
            p for p in chain(*to_visit) if isinstance(p, PermissionAssignment)
        )

        for instance in permissions:
            if (
                instance.permission == permission
                and instance.role == role
                and instance.object == obj
            ):
                return instance

        # Last chance: perform a filtered query. If obj is not None, sometimes
        # getattr(obj, PERMISSIONS_ATTR) has objects not present in session
        # not in this query (maybe in a parent session transaction `new`?).
        if obj is not None and obj.id is None:
            obj = None

        return (
            session.query(PermissionAssignment)
            .filter(
                PermissionAssignment.permission == permission,
                PermissionAssignment.role == role,
                PermissionAssignment.object == obj,
            )
            .first()
        )


class SecurityService(Service):
    name = "security"
    AppStateClass = SecurityServiceState

    def init_app(self, app: Application):
        super().init_app(app)
        state = app.extensions[self.name]
        state.use_cache = True

    def _needs_flush(self):
        """Mark next security queries needs DB flush to have up to date
        information."""
        self.app_state.needs_db_flush = True

    def clear(self):
        pass

    def _current_user_manager(self, session: Session | None = None) -> User:
        """Return the current user, or SYSTEM user."""
        if session is None:
            session = db.session()

        try:
            user = g.user
        except Exception:
            return session.query(User).get(0)

        if sa.orm.object_session(user) is not session:
            # this can happen when called from a task during development
            # (with CELERY_ALWAYS_EAGER=True): the task SA session is not
            # app.db.session, and we should not attach this object to
            # the other session, because it can make weird, hard-to-debug
            # errors related to session.identity_map.
            return session.query(User).get(user.id)
        else:
            return user

    # security log
    @require_flush
    def entries_for(self, obj, limit=20):
        assert isinstance(obj, Entity)
        return (
            SecurityAudit.query.filter(SecurityAudit.object == obj)
            .order_by(SecurityAudit.happened_at.desc())
            .limit(limit)
        )

    # inheritance
    def set_inherit_security(self, obj: FolderishModel, inherit_security: bool):
        assert isinstance(obj, InheritSecurity)
        assert isinstance(obj, Entity)

        obj.inherit_security = inherit_security
        session = object_session(obj) if obj is not None else db.session
        session.add(obj)

        manager = self._current_user_manager(session=session)
        op = (
            SecurityAudit.SET_INHERIT
            if inherit_security
            else SecurityAudit.UNSET_INHERIT
        )
        audit = SecurityAudit(
            manager=manager,
            op=op,
            object=obj,
            object_id=obj.id,
            object_type=obj.entity_type,
            object_name=obj.name,
        )
        session.add(audit)
        self._needs_flush()

    #
    # Roles-related API.
    #
    @require_flush
    def get_roles(
        self,
        principal: AnonymousUser | Group | User,
        object: Model | None = None,
        no_group_roles: bool = False,
    ) -> list[Role]:
        """Get all the roles attached to given `principal`, on a given
        `object`.

        :param principal: a :class:`User` or :class:`Group`

        :param object: an :class:`Entity`

        :param no_group_roles: If `True`, return only direct roles, not roles
        acquired through group membership.
        """
        assert principal
        if hasattr(principal, "is_anonymous") and principal.is_anonymous:
            return [AnonymousRole]

        query = db.session.query(RoleAssignment.role)
        if isinstance(principal, Group):
            filter_principal = RoleAssignment.group == principal
        else:
            filter_principal = RoleAssignment.user == principal
            if not no_group_roles:
                groups = [g.id for g in principal.groups]
                if groups:
                    filter_principal |= RoleAssignment.group_id.in_(groups)

        query = query.filter(filter_principal)

        if object is not None:
            assert isinstance(object, Entity)

        query = query.filter(RoleAssignment.object == object)
        roles = {i[0] for i in query.all()}

        if object is not None:
            for attr, role in (("creator", Creator), ("owner", Owner)):
                if getattr(object, attr) == principal:
                    roles.add(role)
        return list(roles)

    @require_flush
    def get_principals(
        self,
        role: Role,
        anonymous: bool = True,
        users: bool = True,
        groups: bool = True,
        object: Model | None = None,
        as_list: bool = True,
    ) -> Collection[Principal]:
        """Return all users which are assigned given role."""
        if not isinstance(role, Role):
            role = Role(role)
        assert role
        assert users or groups
        query = RoleAssignment.query.filter_by(role=role)

        if not anonymous:
            query = query.filter(RoleAssignment.anonymous == False)
        if not users:
            query = query.filter(RoleAssignment.user == None)
        elif not groups:
            query = query.filter(RoleAssignment.group == None)

        query = query.filter(RoleAssignment.object == object)
        principals = {(ra.user or ra.group) for ra in query.all()}

        if object is not None and role in (Creator, Owner):
            p = object.creator if role == Creator else object.owner
            if p:
                principals.add(p)

        if not as_list:
            return principals

        return list(principals)

    @require_flush
    def _all_roles(self, principal: Principal) -> dict[str | None, set[Role]]:
        query = (
            db.session.query(RoleAssignment.object_id, RoleAssignment.role)
            .outerjoin(Entity)
            .add_columns(Entity._entity_type)
        )

        if isinstance(principal, User):
            filter_cond = RoleAssignment.user == principal
            if principal.groups:
                group_ids = (g.id for g in principal.groups)
                filter_cond |= RoleAssignment.group_id.in_(group_ids)

            query = query.filter(filter_cond)
        else:
            query = query.filter(RoleAssignment.group == principal)

        results = query.all()
        all_roles = {}
        for object_id, role, object_type in results:
            if object_id is None:
                object_key = None
            else:
                object_key = f"{object_type}:{object_id}"
            all_roles.setdefault(object_key, set()).add(role)

        return all_roles

    def _role_cache(self, principal: Principal) -> dict[str | None, set[Role]]:
        if not self._has_role_cache(principal):
            # FIXME: should call _fill_role_cache?
            principal.__roles_cache__ = {}

        return principal.__roles_cache__

    def _has_role_cache(self, principal: Principal) -> bool:
        return hasattr(principal, "__roles_cache__")

    def _set_role_cache(self, principal: Principal, cache: dict[str | None, set[Role]]):
        principal.__roles_cache__ = cache

    def _fill_role_cache(
        self, principal: Principal, overwrite: bool = False
    ) -> dict[str | None, set[Role]]:
        """Fill role cache for `principal` (User or Group), in order to avoid
        too many queries when checking role access with 'has_role'.

        Return role_cache of `principal`
        """
        if not self.app_state.use_cache:
            return {}

        if not self._has_role_cache(principal) or overwrite:
            self._set_role_cache(principal, self._all_roles(principal))
        return self._role_cache(principal)

    @require_flush
    def _fill_role_cache_batch(
        self, principals: Collection[Principal], overwrite: bool = False
    ):
        """Fill role cache for `principals` (Users and/or Groups), in order to
        avoid too many queries when checking role access with 'has_role'."""
        if not self.app_state.use_cache:
            return

        query = db.session.query(RoleAssignment)
        users = {u for u in principals if isinstance(u, User)}
        groups = {g for g in principals if isinstance(g, Group)}
        groups |= {g for u in users for g in u.groups}

        if not overwrite:
            users = {u for u in users if not self._has_role_cache(u)}
            groups = {g for g in groups if not self._has_role_cache(g)}

        if not (users or groups):
            return

        # ensure principals processed here will have role cache. Thus users or
        # groups without any role will have an empty role cache, to avoid
        # unneeded individual DB query when calling self._fill_role_cache(p).
        for p in chain(users, groups):
            self._set_role_cache(p, {})

        filter_cond = []
        if users:
            filter_cond.append(RoleAssignment.user_id.in_(u.id for u in users))
        if groups:
            filter_cond.append(RoleAssignment.group_id.in_(g.id for g in groups))

        query = query.filter(sql.or_(*filter_cond))
        ra_users: dict[User, dict[str, set[Role]]] = {}
        ra_groups: dict[Group, dict[str, set[Role]]] = {}
        for ra in query.all():
            if ra.user:
                all_roles = ra_users.setdefault(ra.user, {})
            else:
                all_roles = ra_groups.setdefault(ra.group, {})

            object_key = (
                f"{ra.object.entity_type}:{ra.object_id:d}"
                if ra.object is not None
                else None
            )
            all_roles.setdefault(object_key, set()).add(ra.role)

        for group, all_roles in ra_groups.items():
            self._set_role_cache(group, all_roles)

        for user, all_roles in ra_users.items():
            for gr in user.groups:
                group_roles = self._fill_role_cache(gr)
                for object_key, roles in group_roles.items():
                    obj_roles = all_roles.setdefault(object_key, set())
                    obj_roles |= roles

            self._set_role_cache(user, all_roles)

    def _clear_role_cache(self, principal: Principal):
        if hasattr(principal, "__roles_cache__"):
            del principal.__roles_cache__

        if isinstance(principal, Group):
            for u in principal.members:
                if hasattr(u, "__roles_cache__"):
                    del u.__roles_cache__

    def has_role(
        self,
        principal: Principal | Role | None,
        role: Collection[Role] | Role | str,
        object: Model | None = None,
    ) -> bool:
        """True if `principal` has `role` (either globally, if `object` is
        None, or on the specific `object`).

        :param:role: can be a list or tuple of strings or a :class:`Role`
            instance

        `object` can be an :class:`Entity`, a string, or `None`.

        Note: we're using a cache for efficiency here. TODO: check that we're not
        over-caching.

        Note2: caching could also be moved upfront to when the user is loaded.
        """
        if not principal:
            return False

        principal = unwrap(principal)
        if not self.running:
            return True

        if isinstance(role, (Role, (str,))):
            role = (role,)

        # admin & manager always have role
        valid_roles = frozenset((Admin, Manager) + tuple(role))

        if AnonymousRole in valid_roles:
            # everybody has the role 'Anonymous'
            return True

        if (
            Authenticated in valid_roles
            and isinstance(principal, User)
            and not principal.is_anonymous
        ):
            return True

        if principal is AnonymousRole or (
            hasattr(principal, "is_anonymous") and principal.is_anonymous
        ):
            # anonymous user, and anonymous role isn't in valid_roles
            return False

        # root always have any role
        if isinstance(principal, User) and principal.id == 0:
            return True

        if object:
            assert isinstance(object, Entity)
            object_key = f"{object.object_type}:{object.id!s}"
            if Creator in role:
                if object.creator == principal:
                    return True
            if Owner in role:
                if object.owner == principal:
                    return True

        else:
            object_key = None

        all_roles = (
            self._fill_role_cache(principal)
            if self.app_state.use_cache
            else self._all_roles(principal)
        )
        roles = set()
        roles |= all_roles.get(None, set())
        roles |= all_roles.get(object_key, set())
        return len(valid_roles & roles) > 0

    def grant_role(
        self, principal: Principal, role: Role | str, obj: Model | None = None
    ):
        """Grant `role` to `user` (either globally, if `obj` is None, or on the
        specific `obj`)."""
        assert principal

        principal = unwrap(principal)
        session = object_session(obj) if obj is not None else db.session
        manager = self._current_user_manager(session=session)
        args = {
            "role": role,
            "object": obj,
            "anonymous": False,
            "user": None,
            "group": None,
        }

        if principal is AnonymousRole or (
            hasattr(principal, "is_anonymous") and principal.is_anonymous
        ):
            args["anonymous"] = True
        elif isinstance(principal, User):
            args["user"] = principal
        else:
            args["group"] = principal

        query = session.query(RoleAssignment)
        if query.filter_by(**args).limit(1).count():
            # role already granted, nothing to do
            return

        # same as above but in current, not yet flushed objects in session. We
        # cannot call flush() in grant_role() since this method may be called a
        # great number of times in the same transaction, and sqlalchemy limits
        # to 100 flushes before triggering a warning
        for ra in (
            o
            for models in (session.new, session.dirty)
            for o in models
            if isinstance(o, RoleAssignment)
        ):
            if all(getattr(ra, attr) == val for attr, val in args.items()):
                return

        ra = RoleAssignment(**args)
        session.add(ra)
        audit = SecurityAudit(manager=manager, op=SecurityAudit.GRANT, **args)
        if obj is not None:
            audit.object_id = obj.id
            audit.object_type = obj.entity_type
            object_name = ""
            for attr_name in ("name", "path", "__path_before_delete"):
                if hasattr(obj, attr_name):
                    object_name = getattr(obj, attr_name)
            audit.object_name = object_name

        session.add(audit)
        self._needs_flush()

        if hasattr(principal, "__roles_cache__"):
            del principal.__roles_cache__

    def ungrant_role(
        self, principal: Principal, role: Role | str, object: Model | None = None
    ):
        """Ungrant `role` to `user` (either globally, if `object` is None, or
        on the specific `object`)."""

        assert principal
        principal = unwrap(principal)
        session = object_session(object) if object is not None else db.session
        manager = self._current_user_manager(session=session)

        args = {
            "role": role,
            "object": object,
            "anonymous": False,
            "user": None,
            "group": None,
        }
        query = session.query(RoleAssignment)
        query = query.filter(
            RoleAssignment.role == role, RoleAssignment.object == object
        )

        if principal is AnonymousRole or (
            hasattr(principal, "is_anonymous") and principal.is_anonymous
        ):
            args["anonymous"] = True
            query.filter(
                RoleAssignment.anonymous == False,
                RoleAssignment.user == None,
                RoleAssignment.group == None,
            )

        elif isinstance(principal, User):
            args["user"] = principal
            query = query.filter(RoleAssignment.user == principal)
        else:
            args["group"] = principal
            query = query.filter(RoleAssignment.group == principal)

        ra = query.one()
        session.delete(ra)
        audit = SecurityAudit(manager=manager, op=SecurityAudit.REVOKE, **args)
        session.add(audit)
        self._needs_flush()
        self._clear_role_cache(principal)

    @require_flush
    def get_role_assignements(self, obj: Model) -> list:
        session = object_session(obj) if obj is not None else db.session
        if not session:
            session = db.session()
        query = session.query(RoleAssignment)
        query = query.filter(RoleAssignment.object == obj).options(
            subqueryload("user.groups")
        )

        role_assignments = query.all()

        results = []
        for ra in role_assignments:
            if ra.anonymous:
                principal = AnonymousRole
            elif ra.user:
                principal = ra.user
            else:
                principal = ra.group
            results.append((principal, ra.role))
        return results

    #
    # Permission API, currently hardcoded
    #
    def has_permission(
        self,
        user: User,
        permission: Permission | str,
        obj: Model | None = None,
        inherit: bool = False,
        roles: None | Role | str | list[Role | str] = None,
    ) -> bool:
        """
        :param obj: target object to check permissions.
        :param inherit: check with permission inheritance. By default, check only
                        local roles.
        :param roles: additional valid role or iterable of roles having
                      `permission`.
        """
        if not isinstance(permission, Permission):
            assert permission in PERMISSIONS
            permission = Permission(permission)
        user = unwrap(user)

        if not self.running:
            return True

        session = None
        if obj is not None:
            session = object_session(obj)

        if session is None:
            session = db.session()

        # root always have any permission
        if isinstance(user, User) and user.id == 0:
            return True

        # valid roles
        # 1: from database
        pa_filter = PermissionAssignment.object == None
        if obj is not None and obj.id is not None:
            pa_filter |= PermissionAssignment.object == obj

        pa_filter &= PermissionAssignment.permission == permission
        valid_roles = session.query(PermissionAssignment.role).filter(pa_filter)
        valid_roles = {res[0] for res in valid_roles.yield_per(1000)}

        # complete with defaults
        valid_roles |= {Admin}  # always have all permissions
        valid_roles |= DEFAULT_PERMISSION_ROLE.get(permission, set())

        # FIXME: obj.__class__ could define default permisssion matrix too

        if roles is not None:
            if isinstance(roles, (Role, str)):
                roles = (roles,)

            for r in roles:
                valid_roles.add(Role(r))

        # FIXME: query permission_role: global and on object

        if AnonymousRole in valid_roles:
            return True

        if Authenticated in valid_roles and not user.is_anonymous:
            return True

        # first test global roles, then object local roles
        checked_objs = [None, obj]
        if inherit and obj is not None:
            while obj.inherit_security and obj.parent is not None:
                obj = obj.parent
                checked_objs.append(obj)

        principals = [user] + list(user.groups)
        self._fill_role_cache_batch(principals)

        return any(
            self.has_role(principal, valid_roles, item)
            for principal in principals
            for item in checked_objs
        )

    def query_entity_with_permission(
        self,
        permission: Permission,
        user: User | None = None,
        Model: type[Model] = Entity,
    ) -> Exists:
        """Filter a query on an :class:`Entity` or on of its subclasses.

        Usage::

            read_q = security.query_entity_with_permission(READ, Model=MyModel)
            MyModel.query.filter(read_q)

        It should always be placed before any `.join()` happens in the query; else
        sqlalchemy might join to the "wrong" entity table when joining to other
        :class:`Entity`.

        :param user: user to filter for. Default: `current_user`.

        :param permission: required :class:`Permission`

        :param Model: An :class:`Entity` based class. Useful when there is more than
        one Entity based object in query, or if an alias should be used.

        :returns: a `sqlalchemy.sql.exists()` expression.
        """
        assert isinstance(permission, Permission)
        assert issubclass(Model, Entity)
        RA = sa.orm.aliased(RoleAssignment)
        PA = sa.orm.aliased(PermissionAssignment)
        # id column from entity table. Model.id would refer to 'model' table.
        # this allows the DB to use indexes / foreign key constraints.
        id_column = sa.inspect(Model).primary_key[0]
        creator = Model.creator
        owner = Model.owner

        if not self.running:
            return sa.sql.exists([1])

        if user is None:
            user = unwrap(current_user)

        # build role CTE
        principal_filter = RA.anonymous == True

        if not user.is_anonymous:
            principal_filter |= RA.user == user

        if user.groups:
            principal_filter |= RA.group_id.in_([g.id for g in user.groups])

        RA = sa.sql.select([RA], principal_filter).cte()
        permission_exists = sa.sql.exists([1]).where(
            sa.sql.and_(
                PA.permission == permission,
                PA.object_id == id_column,
                (RA.c.role == PA.role) | (PA.role == AnonymousRole),
                (RA.c.object_id == PA.object_id) | (RA.c.object_id == None),
            )
        )

        # is_admin: self-explanatory. It search for local or global admin
        # role, but PermissionAssignment is not involved, thus it can match on
        # entities that don't have *any permission assignment*, whereas previous
        # expressions cannot.
        is_admin = sa.sql.exists([1]).where(
            sa.sql.and_(
                RA.c.role == Admin,
                (RA.c.object_id == id_column) | (RA.c.object_id == None),
                principal_filter,
            )
        )

        filter_expr = permission_exists | is_admin

        if user and not user.is_anonymous:
            is_owner_or_creator = sa.sql.exists([1]).where(
                sa.sql.and_(
                    PA.permission == permission,
                    PA.object_id == id_column,
                    sa.sql.or_(
                        (PA.role == Owner) & (owner == user),
                        (PA.role == Creator) & (creator == user),
                    ),
                )
            )
            filter_expr |= is_owner_or_creator

        return filter_expr

    def get_permissions_assignments(
        self, obj: Model | None = None, permission: Permission | None = None
    ) -> dict[Permission, set[Role]]:
        """
        :param permission: return only roles having this permission

        :returns: an dict where keys are `permissions` and values `roles` iterable.
        """
        session = None
        if obj is not None:
            assert isinstance(obj, Entity)
            session = object_session(obj)

            if obj.id is None:
                obj = None

        if session is None:
            session = db.session()

        pa = session.query(
            PermissionAssignment.permission, PermissionAssignment.role
        ).filter(PermissionAssignment.object == obj)

        if permission:
            pa = pa.filter(PermissionAssignment.permission == permission)

        results: dict[permission, set[Role]] = {}
        for perm, role in pa.yield_per(1000):
            results.setdefault(perm, set()).add(role)

        return results

    def add_permission(
        self, permission: Permission, role: Role, obj: Model | None = None
    ):
        session = None
        if obj is not None:
            session = object_session(obj)

        if session is None:
            session = db.session()

        pa = query_pa_no_flush(session, permission, role, obj)

        if not pa:
            pa = PermissionAssignment(permission=permission, role=role, object=obj)

        # do it in any case: it could have been found in session.deleted
        session.add(pa)

    def delete_permission(
        self, permission: Permission, role: Role, obj: Model | None = None
    ):
        session = None
        if obj is not None:
            session = object_session(obj)

        if session is None:
            session = db.session()

        pa = query_pa_no_flush(session, permission, role, obj)

        if pa:
            session.delete(pa)
            if obj:
                # this seems to be required with sqlalchemy > 0.9
                session.expire(obj, [PERMISSIONS_ATTR])

    def filter_with_permission(
        self,
        user: User,
        permission: Permission | str,
        obj_list: list[Model],
        inherit=False,
    ):
        user = unwrap(user)
        return [
            obj
            for obj in obj_list
            if self.has_permission(user, permission, obj, inherit)
        ]


# Instanciate the service
security = SecurityService()



from __future__ import annotations

from contextlib import suppress
from functools import lru_cache
from importlib import resources as rso

import maxminddb

GEOIP_DB_PATH = str(rso.files("abilian.web.admin.panels.geoip") / "ip_country.mmdb")
GEOIP_DB: list[maxminddb.Reader] = []


@lru_cache(maxsize=1024)
def ip_to_country_code(ip: str) -> str:
    """Return 2 letters country code from IP, or empty string if not found."""

    with suppress(OSError, ValueError):
        if not GEOIP_DB:
            GEOIP_DB.append(maxminddb.open_database(GEOIP_DB_PATH))
        result = GEOIP_DB[0].get(ip)
        if result:
            return result.get("country", "")
    return ""



"""Indexing related utilities for Folder, Documents."""

from __future__ import annotations

import sqlalchemy as sa
import sqlalchemy.orm

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.sbe.apps.documents.models import Folder
from abilian.services import get_service

from .models import CmisObject


def reindex_tree(obj: Folder) -> None:
    """Schedule reindexing `obj` and all of its descendants.

    Generally needed to update indexed security.
    """
    assert isinstance(obj, CmisObject)

    index_service = get_service("indexing")
    if not index_service.running:
        return

    descendants = (
        sa.select([CmisObject.id, CmisObject._parent_id])
        .where(CmisObject._parent_id == sa.bindparam("ancestor_id"))
        .cte(name="descendants", recursive=True)
    )
    da = descendants.alias()
    CA = sa.orm.aliased(CmisObject)
    d_ids = sa.select([CA.id, CA._parent_id])
    descendants = descendants.union_all(d_ids.where(CA._parent_id == da.c.id))
    session = sa.orm.object_session(obj) or db.session()

    # including ancestor_id in entity_ids_q will garantee at least 1 value for the
    # "IN" predicate; otherwise when using sqlite (as during tests...)
    # 'ResourceClosedError' will be raised.
    #
    # as an added bonus, "obj" will also be in query results, thus will be added
    # in "to_update" without needing to do it apart.
    entity_ids_q = sa.union(
        sa.select([descendants.c.id]), sa.select([sa.bindparam("ancestor_id")])
    )
    query = (
        session.query(Entity)
        .filter(Entity.id.in_(entity_ids_q))
        .options(sa.orm.noload("*"))
        .params(ancestor_id=obj.id)
    )

    to_update = index_service.app_state.to_update
    key = "changed"

    for item in query.yield_per(1000):
        to_update.append((key, item))



"""A few utility functions.

See https://docs.djangoproject.com/en/dev/topics/http/shortcuts/ for
more ideas of stuff to implement.
"""

from __future__ import annotations

from typing import Any

from flask import current_app
from flask import url_for as flask_url_for
from flask.helpers import send_from_directory
from werkzeug.routing import BuildError


def url_for(obj: Any, **kw: Any) -> str:
    """Polymorphic variant of Flask's `url_for` function.

    Behaves like the original function when the first argument is a
    string. When it's an object, it
    """
    if isinstance(obj, str):
        return flask_url_for(obj, **kw)

    try:
        return current_app.default_view.url_for(obj, **kw)
    except KeyError:
        if hasattr(obj, "_url"):
            return obj._url
        elif hasattr(obj, "url"):
            return obj.url

    raise BuildError(repr(obj), kw, "GET")


def get_object_or_404(cls, *args):
    """Shorthand similar to Django's `get_object_or_404`."""

    return cls.query.filter(*args).first_or_404()


def send_file_from_directory(filename, directory):
    """Helper to add static rules, like in `abilian.app`.app.

    Example use::

       app.add_url_rule(
          app.static_url_path + '/abilian/<path:filename>',
          endpoint='abilian_static',
          view_func=partial(send_file_from_directory,
                            directory='/path/to/static/files/dir'))
    """
    return send_from_directory(directory, filename)
